/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/
#pragma once
#ifndef DATAFOX_COMMUNICATION_PROTOCOL_BASE_H
#define DATAFOX_COMMUNICATION_PROTOCOL_BASE_H

// Basisklasse einbinden.
#include "datafox/communication/connection_base.h"

#include <array>
#include <thread>
#include <mutex>

#include <time.h>
#include <string>

#include <functional>

#include "common/crc.h"
#include "file/interface/studio_setup_structs.h"
#include "datafox/communication/protocol_docking_station_v1.h"
#include "datafox/communication/aes_crypt.h"

///////////////////////////////////////
// Defines zum Paket

// Makros fuer den Zugriff auf die Paketdaten
#define IV_PACKET_START(x) x[0]
#define IV_PACKET_DEVICEID(x) x[1]
#define IV_PACKET_DEVICEID_REMOTE(x) x[2]
#define IV_PACKET_SIZE_CONTROLBLOCK(x) x[3]
#define IV_PACKET_SIZE(x) (*(uint16_t *)&x[4])
#define IV_PACKET_FLAGS(x) x[6]
#define IV_PACKET_RESERVE(x) x[7]
#define IV_PACKET_CRC_HEAD(x) (*(uint16_t *)&x[8])

#define IV_PACKET_COMMAND(x) x[IV_PACKET_LENGTH_HEAD]
#define IV_PACKET_MODULE(x) (*(uint16_t *)&x[IV_PACKET_LENGTH_HEAD + IV_PACKET_LENGTH_CONTROLBLOCK + sizeof( uint8_t )])
#define IV_PACKET_FUNCTION(x) (*(uint16_t *)&x[IV_PACKET_LENGTH_HEAD + IV_PACKET_LENGTH_CONTROLBLOCK + sizeof( uint8_t ) + sizeof( uint16_t )])

#define IV_PACKET_CRC(x) (*(uint16_t *)&x[IV_PACKET_LENGTH(x) - sizeof( uint16_t )])

// Grundangaben
#if _DEBUG && 0
#define IV_PACKET_LENGTH_MAX                0x580 // Maximalgroesse fuer ein Paketspeicher
#else
#define IV_PACKET_LENGTH_MAX                0x300 // Maximalgroesse fuer ein Paketspeicher
#endif
#define IV_PACKET_LENGTH_HEAD               10    // Laenge des Paketkopfes
#define IV_PACKET_LENGTH_CONTROLBLOCK        1    // Laenge der Steuerinformation. uint8_t(Packetnummer)
#define IV_PACKET_LENGTH_FRAME              12    // Minimalframegroesse (Paketkopf + 16bit CRC)
#define IV_PACKET_LENGTH_DATA               (IV_PACKET_LENGTH_MAX - IV_PACKET_LENGTH_FRAME - IV_PACKET_LENGTH_CONTROLBLOCK)
#define IV_PACKET_LENGTH_COMMANDBLOCK        5    // uint8_t(Packetcommando, version + anfrage|antwort), uint16_t(Modul), uint16_t(Funktion)
#define IV_PACKET_LENGTH_COMMAND_PARAMS     (IV_PACKET_LENGTH_DATA - IV_PACKET_LENGTH_COMMANDBLOCK)

// Paketsonderzeichen 
#define IV_PACKET_SIGN_START                0x7E  // ~ als Startzeichen der Pakete Byte 1
#define IV_PACKET_SIGN_ESCAPE               0x7F  // als Escapezeichen der Pakete

// Paketflags Byte 6
// Achtung: !!! Die unteren 3 bit geben einen Wert an und sind somit keine seperaten Flags!!!
#define IV_PACKET_CRYPT_MODE_NA             0x00 // Not available. Keine Unterstuetzung von verschluesselter Kommunikation.
#define IV_PACKET_CRYPT_MODE_NO             0x01 // No. Verschluesselte Kommunikation wird unterstuetzt muss aber unverschluesselt kommunizieren.
#define IV_PACKET_CRYPT_MODE_DEFAULT        0x02 // Default. Verschluesselung mit dem Datafox-Standard-Schluessel.
#define IV_PACKET_CRYPT_MODE_PASSPHRASE     0x03 // Passphrase. Es muss mit einem Benutzerdefinierten Passwort verschluesselt werden.
// Dieses hier sind nun Flags.
#define IV_PACKET_FLAG_CRYPTED              0x04 // Dateninhalt liegt verschluesselt vor.
#define IV_PACKET_FLAG_RESERVED1            0x08 // 
#define IV_PACKET_FLAG_FLASH_SERVICE        0x10 // Flash Service: Globaler Wear-Level ist >= 6				
#define IV_PACKET_FLAG_FLASH_WARNING        0x20 // Flash Warning: CRC Fehler im Ringspeicher oder Listenbereich. (Setup und Listen neu übertragen)		
#define IV_PACKET_FLAG_FLASH_LOCKED         0x40 // Flash Error: Keine freien Seiten mehr verfügbar, Gerät befindet sich im Bedienstopp.
#define IV_PACKET_FLAG_RESERVED2            0x80 // 
///////////////////
// Paketdaten

// Paketkommando (unteres Nibble)
#define IV_PACKET_VERSION_4                 0x04 // Paketversion 4 seit FW 04.01.xx und Bootloader 04.xx
// Paketkommando (oberes Nibble)
#define IV_PACKET_ANSWER                    0x80 // Paket ist eine Antwort
#define IV_PACKET_REQUEST                   0x40 // Paket ist eine Anfrage
#define IV_PACKET_ERROR                     0x20 // Paket transportiert einen Fehler (Anfrage oder Antwort)


/////////////////////////////
// Werte fuer die Umgebung

// Standardwerte als Vorgabe
#define IV_DEFAULT_DEVICEID_STANDALONE      0xFE // Modbusadresse fuer Direktverbindung

#define IV_PENDING_TIMEOUT                  60000 // Timeout fuer eine Aktive Antwort (ab 4.1.5.9 von 15 auf 25s, ab 4.2.3 auf 60s)
#define IV_PENDING_FINGERPRINT_TIMEOUT      60000 // Timeout fuer eine Aktive Antwort (ab 4.2.1, ab 4.2.3 von 50 auf 60s)
#define IV_PENDING_FINGERPRINT_GERNERATE_TIMEOUT      300000 // Timeout fuer eine Aktive Antwort (ab 4.3.10 auf 300s)

#define IV_PENDING_RESET_TIMEOUT            95000 // Timeout fuer eine Aktive Antwort bei SYSTEM.SYSTEMRESET (ab 4.2.1, ab 4.3.3 von 5000 auf 95s)
#define IV_PENDING_TRIALS(to)               (!(IV_PENDING_TIMEOUT % to) ? IV_PENDING_TIMEOUT / to : IV_PENDING_TIMEOUT / to + 1)
#define IV_PENDING_FINGERPRINT_TRIALS(to)   (!(IV_PENDING_FINGERPRINT_TIMEOUT % to) ? IV_PENDING_FINGERPRINT_TIMEOUT / to : IV_PENDING_FINGERPRINT_TIMEOUT / to + 1)
#define IV_PENDING_FINGERPRINT_GENERATE_TRIALS(to)   (!(IV_PENDING_FINGERPRINT_GERNERATE_TIMEOUT % to) ? IV_PENDING_FINGERPRINT_GERNERATE_TIMEOUT / to : IV_PENDING_FINGERPRINT_GERNERATE_TIMEOUT / to + 1)
#define IV_PENDING_RESET_TRIALS(to)         (!(IV_PENDING_RESET_TIMEOUT % to) ? IV_PENDING_RESET_TIMEOUT / to : IV_PENDING_RESET_TIMEOUT / to + 1)
#define IV_SKIPCOUNT                        1024 // Maximale Anzahl Zeichen die auserhalbt des Packetzusammenhangs zu lesen sind.

// Contexte fuer call-Routine
#define IV_CONTEXT_MASTER                    0x0000 // Ausfuehrung als Kommunikationsbefehl
#define IV_CONTEXT_SLAVE                     0x0001 // Pruefen auf anstehende Kommunikationsbefehle der Gegenstelle
#define IV_CONTEXT_TRANSMIT_ANSWER           0x0002 // Senden einer Antwort
#define IV_CONTEXT_TRANSMIT_ANSWER_ERROR     0x0004 // Senden einer Antwort als Fehler

// Maximale Laenge eines Fehlertextes
#define ERROR_TEXT_SIZE                      255

// Auswertung der Fehlerbestandteile
#define IV_RESULT_MODULE(res)                ((res >> 8) & 0xFF)
#define IV_RESULT_FUNCTION(res)              (res & 0xFF)

///////////////////////////////////////
// Modul und Funktionsbasis

/** In diesem Module sind alle Kommunikationskontroll Funktionen enthalten.
*
* Es handelt sich hierbei um ein Modul zur Kontrolle der auszufuehrenden Funktionen
* der abgeleiteten Module.
*
*/
#define IV_MODULE_BASE                       0xFFFF
// Undefinierte Modulfunktion
#define IV_FUNCTION_BASE                     0xFFFF

/** In diesem Module sind alle Protokolspezifischen Funktionen enthalten.
*
* 
*
*/
#define IV_MODULE_PROTOCOL                       0x0000

// Prueft die Erreichbarkeit eines Geraets. Es wird nur ein einzelnes Paket versendet.
#define IV_FUNCTION_PING                         0x0000

namespace Datafox 
{
    namespace Communication 
    {
        /// Basisklasse fuer pure protokollbasierte Kommunikationsaufgaben.
        class ProtocolBase
        {
        public:
            // Standardwerte fuer Parameter
            static constexpr uint8_t const defaultDeviceAddress()
            {
                return 0;
            }
            static constexpr uint8_t const defaultDeviceNumberStandAlone()
            {
                return 254;
            }

        public:
            ProtocolBase( int32_t connectionNumber = ConnectionBase::defaultConnectionNumber(), uint8_t deviceAddress = defaultDeviceAddress() );
            virtual ~ProtocolBase();

        public:
            inline uint8_t lastFlags()
            {
                return m_lastFlags;
            }
            inline void setLastFlags( uint8_t flags )
            {
                m_lastFlags = flags;
            }
            // Speicher fuer Pakete/Daten
        public:
            uint8_t m_packetSend[IV_PACKET_LENGTH_MAX];
            uint8_t m_packetReceive[IV_PACKET_LENGTH_MAX];
            uint8_t m_temp[IV_PACKET_LENGTH_MAX * 2];
            uint8_t *m_tempPos;

            void reset();

        private:
            bool m_packetStart;
            uint32_t m_skipCount;
            uint8_t m_packetNumber; // Paketzaehler
        protected:
            uint8_t m_lastFlags;    // Halten der letzten empfangenen Flagvariable aus Paket. Für Flash-Service Flags etc...
        private:
            double m_qs;

            clock_t m_clockWrite;    // Dauer des letzten Versendevorgangs
            clock_t m_clockRead;     // Dauer des letzten Empfangvorgangs
            clock_t m_clocksWrite;   // Gesamtdauer der Versendevorgaenge
            clock_t m_clocksRead;    // Gesamtdauer der Empfangsvorgaenge
            uint32_t m_packetWriteCounter;       // Anzahl gueltig versendeter Pakete
            uint32_t m_packetWritePayloadSize;   // Anzahl der versendeten Nutzdaten-Bytes
            uint32_t m_packetReadCounter;        // Anzahl gueltig empfangener Pakete
            uint32_t m_packetLostCounter;        // Anzahl unbeantworteter Pakete
            uint32_t m_packetReadPayloadSize;    // Anzahl der gelesenen Nutzdaten-Bytes

            uint8_t m_deviceNumberSender;   // Die zu verwendende deviceAddress der Klasse
            uint8_t m_deviceNumberReceiver; // Die zu verwendete deviceAddress des Remote der Klasse
        protected:
            // Membervariablen fuer call-Routine
            uint16_t m_previousModule;   // Zuletzt ausgefuehrtes Modul
            uint16_t m_previousFunction; // Zuletzt ausgefuehrte Funktion

            // Routinen fuer device_address's
        public:
            bool isOwnDeviceNumber( uint8_t deviceAddress )
            {
                return deviceAddress == deviceNumberSender();
            }

            uint8_t deviceNumberSender()
            { 
                return m_deviceNumberSender; 
            }
            uint8_t setDeviceNumberSender( uint8_t deviceAddress )
            { 
                uint8_t device_address_previous = m_deviceNumberSender;
                m_deviceNumberSender = deviceAddress; 
                return device_address_previous; 
            }
            uint8_t deviceNumberReceiver()
            { 
                return m_deviceNumberReceiver; 
            }
            uint8_t setDeviceNumberReceiver( uint8_t deviceAddress )
            { 
                uint8_t device_address_previous = m_deviceNumberReceiver;
                m_deviceNumberReceiver = deviceAddress; 
                return device_address_previous; 
            }

            // Routinen zur Ermittlung von sonstigen Memberwerten
        public:
            inline void setPacketNumber( uint8_t number )
            {
                m_packetNumber = number;
            }
            inline void incPacketNumber()
            {
                ++m_packetNumber;
            }
            inline void decPacketNumber()
            {
                --m_packetNumber;
            }
            inline uint8_t packetNumber()
            {
                return m_packetNumber;
            }

            // Standardroutinen
        public:
            // Call-Back-Funktion
            std::function<int32_t()> m_callBackFunction;
        public:
            void setCallBackFunction( std::function<int32_t()> func );

            // Routinen fuer CRC 16 sowie 32bit
        private:
            uint32_t m_crc32Table[256]; // Lookup table array
            void initCrc32Table();
            uint32_t reflect( uint32_t ref, int8_t ch );
        public:
            uint32_t crc32( uint8_t *data, std::size_t size, uint32_t crc = 0xFFFFFFFF, bool closure = true );
            uint16_t crc16( uint8_t *data, std::size_t size, uint16_t crc = 0xFFFF );

            // Routinen fuer Verschluesselung der Paketdaten
        private:
            uint8_t m_sbox[256];
            uint8_t m_k[256];
            uint8_t m_key[32];
        public:
            // Verschluesselungsroutinen
            void df4LoadCryptKey();
            void df4Init( bool binary_key );
            void df4( uint8_t *data, std::size_t size );

        protected:
            time_t m_timeReadPacket;  // Zeitpunkt des letzten Paketempfangs
        public:
            time_t timeReadPacket()
            {
                return m_timeReadPacket;
            }
        protected:
            time_t m_timeWritePacket; // Zeitpunkt des letzten Paketversands
        public:
            time_t timeWritePacket()
            {
                return m_timeWritePacket;
            }
        protected:
            // Paketdaten ggf. verschluesseln.
            virtual int writeProtocolCrypt()
            { 
                return IV_ERROR_SUCCESS; 
            }
            // Packetdaten Byteweise schreiben und dabei Maskieren, komplettpacket senden.
            int32_t writeProtocol( uint8_t *buffer, uint8_t writeTrials = ConnectionBase::defaultWriteTrials() );
            // Aufbereiten und senden eines Protokollpakets
            int32_t writePacket( uint8_t deviceAddress, uint8_t type, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t count, uint8_t writeTrials = ConnectionBase::defaultWriteTrials() );
            // Zu versendende Daten ggf. verschluesseln.
            virtual int32_t writeStreamCrypt( std::string &buffer )
            { 
                (void)buffer; 
                return IV_ERROR_SUCCESS;
            }
            // Zu empfangende Daten ggf. entschluesseln.
            virtual int32_t readStreamCrypt( std::string &buffer )
            { 
                (void)buffer; 
                return IV_ERROR_SUCCESS;
            }
            // Ermitteln ob akutell eine Verschluesselungssession aktiv ist.
            virtual bool isCrypt()
            { 
                return false;
            }

            // Paketdaten Byteweise empfangen und Demaskierung durchfuehren.
            int32_t readProtocol( uint8_t *buffer, uint8_t readTrials = ConnectionBase::defaultReadTrials() );
            int32_t readTimeboyStation( uint8_t *buffer, uint8_t readTrials = ConnectionBase::defaultReadTrials() );
            // Packetdaten ggf. entschluesseln
            virtual int32_t readProtocolCrypt()
            { 
                return IV_ERROR_SUCCESS;
            }
            // Empfang und aufbereiten eines Protokollpakets
            int32_t readPacket( uint8_t deviceAddress, uint8_t &type, uint16_t &module, uint16_t &func, uint8_t *buffer, uint16_t &count, uint8_t readTrials = ConnectionBase::defaultReadTrials() );

            // Daten eines Packets raustracen
            void tracePacket( uint8_t *packetBuffer, bool crlf = true );

        public:
            // Verwaltungsklasse fuer Dockingstation im Channel
            ProtocolDockingstationV1 m_timeboyStation;
            // Ermittlung ob es sich um einen Slot einer TimeboyStation handelt.
            int isTimeboySlot( uint8_t deviceAddress )
            {
                int32_t result = 0;
                if ( m_timeboyStationEnabled )
                {
                    result = m_timeboyStation.isSlot( deviceAddress );
                }

                return result;
            }
            // Beginn und Ende der Kommunikation ueber einen TimeboySlot
            virtual int32_t setTimeboySlotState( uint8_t device_address, int32_t state, int32_t commandResult = IV_ERROR_SUCCESS )
            {
                int32_t result = -IV_ERROR_SLOT_NONE;
                if ( m_timeboyStationEnabled )
                {
                    result = m_timeboyStation.setTimeboySlotState( device_address, state, commandResult );
                }

                return result;
            }
            void resetTimeboyStation()
            {
                if ( m_timeboyStationEnabled )
                {
                    m_timeboyStation.resetSlots();
                }
            }
            void setTimeboyStationEnabled( bool enabled )
            {
                m_timeboyStationEnabled = enabled;
            }
        protected:
            uint32_t m_timeboyBaseAddress;
            bool m_timeboyStationEnabled;
 
        public:
            // Ausfuehren eines Protokollbefehls
            virtual int32_t call( uint16_t context, uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count, int32_t contextValue = -1 );
            std::function<int32_t ( uint16_t module, uint16_t func, uint8_t *buffer, uint16_t &count)> remoteCall;

        protected:
            std::shared_ptr<ConnectionBase> m_connection;

        public:
            std::shared_ptr<ConnectionBase> connectionPointer()
            {
                return m_connection;
            }
            void setConnectionPointer( std::shared_ptr<ConnectionBase> base )
            {
                m_connection = base;
            }
            int32_t connectionNumber()
            {
                return connection().number();
            }

        public:
            // Gibt die Grundklasse fuer pure Kommunikationsaufgaben zurueck
            ConnectionBase& connection() 
            { 
                return *m_connection.get();
            }

        private:
            std::shared_ptr<TraceBase> m_trace;

        public:
            std::shared_ptr<TraceBase> tracePointer()
            {
                return m_trace;
            }
            void setTracePointer( std::shared_ptr<TraceBase> base )
            {
                m_trace = base;
            }

        public:
            virtual void trace( const char *format... )
            {
                if ( m_trace.get() != nullptr && m_trace->mode() >= TraceBase::TraceMode::on )
                {
                    va_list args;
                    va_start( args, format );
                    m_trace->print( format, args );
                    va_end( args );
                }
            }

            TraceBase::TraceMode traceMode()
            {
                TraceBase::TraceMode mode = TraceBase::TraceMode::off;
                if ( m_trace.get() != nullptr )
                {
                    mode = m_trace->mode();
                }

                return mode;
            }
            void setTraceMode( TraceBase::TraceMode mode = TraceBase::TraceMode::on )
            {
                if ( m_trace.get() != nullptr )
                {
                    m_trace->setMode( mode );
                }
            }

            std::string traceFilePath()
            {
                std::string filePath;
                if ( m_trace.get() != nullptr )
                {
                    filePath = m_trace->filePath();
                }

                return filePath;
            }
            void setTraceFilePath( const std::string &file_path )
            {
                if ( m_trace.get() != nullptr )
                {
                    m_trace->setFilePath( file_path );
                }
            }

            // Fehlerbearbeitung
        public:
            void resetError();

            // Fehlertexte fuer die Kommunikation/Module/Funktionen
            uint16_t m_errorCommunicationNumber;
            std::string m_errorCommunicationText;
            void errorCommunication( uint16_t error, std::string &text );
        };

    } // namespace Communication
} // namespace Datafox

#endif
