/**
* @file  file_hardware_info.cpp
* @brief Implementierung der Klasse FileHardwareInfo.
*/
//////////////////////////////////////////////////////////////////////

#include "file/file_hardware_info.h"
#include <memory.h>
#include <stdio.h>

#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <boost/format.hpp>


/*! Schnittstelle zwischen der Datei mit den Daten für die HardwareConfig und den Softwaremodulen.
Die Daten werden wegen ihrer Baumartigen Art nach dem Laden der Datei in eine Baumstruktur von
ItemHardwareInfo Einträgen gewandelt. Änderungen bitte in diesem Item-Baum vornehmen, über 
rootItemHardwareConfig() erhält man zugriff auf das Wurzelitem.
Die Baumstruktur wird vor dem Speichern serialisiert und in der Datei gespeichert.
*/
FileHardwareInfo::FileHardwareInfo()
{
    // Reset der sonstigen Umgebung durchführen.
    reset();
}

FileHardwareInfo::FileHardwareInfo( const FileHardwareInfo &hwInfo )
{
    // Klasse initialisieren.
    reset();

    m_file_path = hwInfo.m_file_path;
    _crc = hwInfo._crc;
    m_buffer = hwInfo.m_buffer;

    // Member durch gelesene Daten initialisieren und pruefen
    if ( m_buffer.size() >= (sizeof(HARDWARECONFIGNODE) + sizeof(unsigned int) + sizeof(HARDWARECONFIG)) )
    {
        // Pruefsumme zum Vergleich berechnen.
        if ( _crc != checksum32bit() )
        {
            // Angegebene Prüfsumme stimmt nicht mit der berechneten überein.
        }

        HARDWARECONFIGNODE *hcn = (HARDWARECONFIGNODE*)&m_buffer.front();
        if ( hcn->version != 1 )
        {
            // Version der Dateidaten aktueller als erwartet.
        }

        // Aus den Serialisierungsdaten die Baumrepresentation erstellen.
        if ( !readSerialize() )
        {
            // Erstellen der Baumdaten gescheitert.
        }
    }
    else
    {
        // Keine Daten vorhanden
    }
}

FileHardwareInfo::~FileHardwareInfo()
{
}

FileHardwareInfo &FileHardwareInfo::operator=( const FileHardwareInfo &hwInfo )
{
    // Klasse initalisieren
    reset();

    m_file_path = hwInfo.m_file_path;
    _crc = hwInfo._crc;
    m_buffer = hwInfo.m_buffer;

    // Member durch gelesene Daten initialisieren und pruefen
    if ( m_buffer.size() >= (sizeof(HARDWARECONFIGNODE) + sizeof(unsigned int) + sizeof(HARDWARECONFIG)) )
    {
        // Pruefsumme zum Vergleich berechnen.
        if ( _crc != checksum32bit() )
        {
            // Angegebene Prüfsumme stimmt nicht mit der berechneten überein.
        }

        HARDWARECONFIGNODE *hcn = (HARDWARECONFIGNODE*)&m_buffer.front();
        if ( hcn->version != 1 )
        {
            // Version der Dateidaten aktueller als erwartet.
        }

        // Aus den Serialisierungsdaten die Baumrepresentation erstellen.
        if ( !readSerialize() )
        {
            // Erstellen der Baumdaten gescheitert.
        }
    }
    else
    {
        // Keine Daten vorhanden
    }

    return *this;
}

/*! Initalisiert die Klassendaten.
Nach dem Aufruf stellt die Klasse eine Leere Datei dar.
*/
void FileHardwareInfo::reset()
{
    // Basis Reset
    FileIntelHex::resetData();

    // Member initialisieren

    // Dateiname
    m_file_path.clear();

    // Pruefsumme zum Test auf Datenaenderung ruecksetzen.
    _crc = 0;

    // Signature Call-Back
    signing = 0;

    // Der Standardmäßig zu verarbeitende Datensatztyp
    setRecordType( Record_type::hardware_config );

    // Vorhandene Konfiguration löschen.
    _deviceHardwareInfo.reset();
}

/*! Laden einer HardareConfig-Datei und wandeln der Daten in eine interne 
Baumdarstellung von ItemHardwareInfo-Klassen.
Die Datei muss in einem Intel-Hex-Format vorliegen. Die Endung sollte *.hex oder *.df1 sein.
*/
bool FileHardwareInfo::load( const std::string &filePath ) 
{
    // Daten durch die Basisklasse in _buffer laden.
    if ( !FileIntelHex::load( filePath ) )
        return false;

    // Member durch gelesene Daten initialisieren und pruefen
    if ( m_buffer.size() >= (sizeof(HARDWARECONFIGNODE) + sizeof(unsigned int) + sizeof(HARDWARECONFIG)) )
    {
        // Pruefsumme zum Test auf Datenaenderung berechnen.
        _crc = checksum32bit();

        HARDWARECONFIGNODE *hcn = (HARDWARECONFIGNODE*)&m_buffer.front();
        if ( hcn->version != 1 )
        {
            // Version der Dateidaten aktueller als erwartet.
            setErrorState( Error_state::version );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }

        // Aus den Serialisierungsdaten die Baumrepresentation erstellen.
        return readSerialize();
    }
    else
    {
        // Fehler zuweisen
        setErrorState( Error_state::no_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
    }

    return false;
}

/*!
Hilfsfunktion zum manuellen Serialisieren der in _memHardwareConfig enthaltenen Daten.
*/
bool FileHardwareInfo::readSerialize()
{
    // Aus den Serialisierungsdaten die Baumrepresentation erstellen.
    if ( !_deviceHardwareInfo.readSerialize( m_buffer ) )
    {
        // Erstellen der Dateidaten gescheitert
        setErrorState( Error_state::serialize_file_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return false;
    }

    return true;
}

/*! Speichern der internen Darstellung von ItemHardwareInfo-Klassen in eine HardwareConfig-Datei.
Die Datei wird im Intel-Hex-Format gespeichert. Die Endung sollte *.hex oder *.df1 sein.
*/
bool FileHardwareInfo::save( const std::string &filePath, bool appendToFile /*= false*/, bool appendEndLine /*= true */ ) 
{
    // Serialisierung der Baumstruktur in das zu schreibende Array durchfuehren.
    if ( !writeSerialize() )
        return false;

    // Signatur ggf. nachtragen.
    if ( signing )
    {
        HARDWARECONFIGNODE *hcn = (HARDWARECONFIGNODE*)&m_buffer.front();
        hcn->signature = signing( &m_buffer[hcn->ptrHardwareConfigArray], m_buffer.size() - hcn->ptrHardwareConfigArray );
        hcn->signatureSize = static_cast<unsigned int>( m_buffer.size() ) - hcn->ptrHardwareConfigArray;
    }

    // Nur bei einer Dateiangabe auch wirklich in diese schreiben.
    if ( filePath.size() )
    {
        // Daten der HARDWARE-Konfiguration sichern.
        if ( !FileIntelHex::save( filePath, appendToFile, appendEndLine ) )
            return false;
    }

    // Pruefsumme zum Test auf Datenaenderung berechnen.
    _crc = checksum32bit();

    return true;
}

/*!
Hilfsfunktion zum manuellen Serialisieren der in _memHardwareConfig enthaltenen Daten.
*/
bool FileHardwareInfo::writeSerialize()
{
    // Aktuellen Speicher leeren.
    m_buffer.clear();
    // Serialisierung der Baumstruktur in das zu schreibende Array durchfuehren.
    if ( !_deviceHardwareInfo.writeSerialize( m_buffer ) )
    {
        // Erstellen der Dateidaten gescheitert
        setErrorState( Error_state::serialize_file_data );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return false;
    }

    return true;
}

listItemBoard FileHardwareInfo::collectModuleIds( const setModuleId &ids )
{
    return rootItemDeviceHardwareInfo().collectModuleIds( ids );
}

listItemBoard FileHardwareInfo::moduleRS232Communication( const setModuleRS232Communication &module )
{
    return rootItemDeviceHardwareInfo().moduleRS232Communication( module );
}

listItemBoard FileHardwareInfo::moduleRS485Communication( const setModuleRS485Communication &module )
{
    return rootItemDeviceHardwareInfo().moduleRS485Communication( module );
}

listItemBoard FileHardwareInfo::onBoardRS485Communication( const setOnBoardRS485Communication &module )
{
    return rootItemDeviceHardwareInfo().onBoardRS485Communication( module );
}

listItemBoard FileHardwareInfo::moduleUSBCommunication( const setModuleUSBCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleUSBCommunication( module );
}

listItemBoard FileHardwareInfo::moduleLANCommunication( const setModuleLANCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleLANCommunication( module );
}

listItemBoard FileHardwareInfo::onBoardLANCommunication( const setOnBoardLANCommunication &boards )
{
    return rootItemDeviceHardwareInfo().onBoardLANCommunication( boards );
}

listItemBoard FileHardwareInfo::moduleWLANCommunication( const setModuleWLANCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleWLANCommunication( module );
}

listItemBoard FileHardwareInfo::moduleMobileCommunication( const setModuleMobileCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleMobileCommunication( module );
}

listItemBoard FileHardwareInfo::moduleBluetoothCommunication( const setModuleBluetoothCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleBluetoothCommunication( module );
}

listItemBoard FileHardwareInfo::moduleMMCCommunication( const setModuleMMCCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleMMCCommunication( module );
}

listItemBoard FileHardwareInfo::moduleUSBHostCommunication( const setModuleUSBHostCommunication &module )
{
    return rootItemDeviceHardwareInfo().moduleUSBHostCommunication( module );
}

listItemBoard FileHardwareInfo::moduleAccessControl( const setModuleAccessControl &module )
{
    return rootItemDeviceHardwareInfo().moduleAccessControl( module );
}

listItemBoard FileHardwareInfo::onBoardAccessControl( const setOnBoardAccessControl &boards )
{
    return rootItemDeviceHardwareInfo().onBoardAccessControl( boards );
}

listItemBoard FileHardwareInfo::moduleBiokey( const setModuleBiokey &module )
{
    return rootItemDeviceHardwareInfo().moduleBiokey( module );
}

listItemBoard FileHardwareInfo::moduleGPS( const setModuleGPS &module )
{
    return rootItemDeviceHardwareInfo().moduleGPS( module );
}

listItemBoard FileHardwareInfo::moduleDisplay( const setModuleDisplay &module )
{
    return rootItemDeviceHardwareInfo().moduleDisplay( module );
}

listItemBoard FileHardwareInfo::moduleTouch( const setModuleTouch &module )
{
    return rootItemDeviceHardwareInfo().moduleTouch( module );
}

listItemBoard FileHardwareInfo::moduleTemperatureSensor( const setModuleTemperatureSensor &module )
{
    return rootItemDeviceHardwareInfo().moduleTemperatureSensor( module );
}

/*! Stellt die Automatische Korrektur von Fehlenden Steckmodulen ein während dem Laden der Daten ein.

*/
void FileHardwareInfo::setAutoCorrectModulesDisabled( bool state )
{
    rootItemDeviceHardwareInfo().setAutoCorrectModuleDisabled( state );
}